/*
 * sprintf.c - sprintf for DATA
 *
 * Written 2001,2002 by Werner Almesberger
 * Copyright 2001 EPFL-ICA
 * Copyright 2001,2002 Bivio Networks
 *
 * Based on util.c:alloc_sprintf
 */


#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <assert.h>

#include "util.h"
#include "error.h"
#include "data.h"
#include "sprintf.h"


static DATA next_data(const DATA_LIST **curr)
{
    DATA d;

    if (!*curr) yyerror("not enough arguments for format string");
    d = *(*curr)->ref;
    *curr = (*curr)->next;
    return d;
}


char *sprintf_data(const char *fmt,const DATA_LIST *data_list)
{
    const DATA_LIST *curr;
    const char *s;
    char *buf,*t;
    int total;

    total = 0;
    curr = data_list;
    for (s = fmt; *s; s++)
	if (*s != '%') total++;
	else {
	    int len = 0,min = 0;

	    s++;
	    if (*s == '-' || isdigit(*s)) {
		char *end;

		if (*s == '-') s++;
		min = strtoul(s,&end,10);
		s = end;
	    }
	    if (*s == '*') {
		min = data_convert(next_data(&curr),dt_unum).u.unum;
		if (min < 0) min = -min;
		s++;
	    }
	    if (*s == '.') {
		int real_len;

		if (s[1] != '*' || s[2] != 's')
		    yyerror("%%. format only allowed in %%.*s");
		len = data_convert(next_data(&curr),dt_unum).u.unum;
		real_len =
		  strlen(data_convert(next_data(&curr),dt_string).u.string);
		if (real_len < len) len = real_len;
		s += 2;
	    }
	    else {
		int tmp;

		switch (*s) {
		    case 'c':
			len = 1;
			(void) data_convert(next_data(&curr),dt_unum).u.unum;
			break;
		    case 's':
			len = strlen(
			  data_convert(next_data(&curr),dt_string).u.string);
			break;
		    case 'd':
			tmp = data_convert(next_data(&curr),dt_unum).u.unum;
			/* warn ? */
			if (tmp < 0) {
			    tmp = -tmp;
			    len++;
			}
			len += log_10(tmp);
		        break;
		    case 'u':
			len = log_10(
			  data_convert(next_data(&curr),dt_unum).u.unum);
		        break;
		    case 'x':
		    case 'X':
			len = log_16(
			  data_convert(next_data(&curr),dt_unum).u.unum);
		        break;
		    case '%':
			len = 1;
			break;
		    default:
			yyerrorf("unrecognized format character \"%c\"",*s);
		}
	    }
	    total += len < min ? min : len;
	}
    if (curr)
	yywarnf("ignoring extra sprintf argument%s",curr->next ? "s" : "");
    buf = t = alloc(total+1);
    curr = data_list;
    for (s = fmt; *s; s++)
	if (*s != '%') *t++ = *s;
	else {
	    int min = 0;

	    s++;
	    if (*s == '-' || isdigit(*s)) {
		char *end;

		min = strtol(s,&end,10);
		s = end;
	    }
	    if (*s == '*') {
		min = data_convert(next_data(&curr),dt_unum).u.unum;
		s++;
	    }
	    if (*s == '.') {
		int len;

		if (s[1] != '*' || s[2] != 's') abort();
		len = data_convert(next_data(&curr),dt_unum).u.unum;
		t += sprintf(t,"%*.*s",min,len,
		  data_convert(next_data(&curr),dt_string).u.string);
		s += 2;
	    }
	    else {
		switch (*s) {
		    case 'c':
			t += sprintf(t,"%*c",min,(int)
			  data_convert(next_data(&curr),dt_unum).u.unum);
		        break;
		    case 's':
			t += sprintf(t,"%*s",min,
			  data_convert(next_data(&curr),dt_string).u.string);
			break;
		    case 'd':
			t += sprintf(t,"%*ld",min,(long)
			  data_convert(next_data(&curr),dt_unum).u.unum);
		        break;
		    case 'u':
			t += sprintf(t,"%*lu",min,(unsigned long)
			  data_convert(next_data(&curr),dt_unum).u.unum);
		        break;
		    case 'x':
			t += sprintf(t,"%*lx",min,(unsigned long)
			  data_convert(next_data(&curr),dt_unum).u.unum);
		        break;
		    case 'X':
			t += sprintf(t,"%*lX",min,(unsigned long)
			  data_convert(next_data(&curr),dt_unum).u.unum);
		        break;
		    case '%':
			*t++ = '%';
			break;
		    default:
			abort();
		}
	    }
	}
    assert(t-buf == total);
    *t = 0;
    return buf;
}
